import collections
import csv
import dataclasses
import enum
import fractions
import functools
import importlib.resources
import typing

from elite_engineering import trade


@dataclasses.dataclass
class Material:
    symbol: str
    name: str
    type_: str
    category: str
    grade: int

    def trade_ratio(self, other: "Material") -> typing.Optional[fractions.Fraction]:
        if self.type_ != other.type_:
            return None
        return trade.ratio(self.grade, other.grade, self.category != other.category)


def _load() -> typing.Dict[str, Material]:
    path = importlib.resources.files(__package__).joinpath("data/fdevids/material.csv")
    lines = path.read_text().splitlines()
    reader = csv.reader(lines[1:])
    return {
        symbol.lower(): Material(symbol.lower(), name, type_, category, int(grade))
        for fdevid, symbol, grade, type_, category, name in reader
    }


materials: typing.Dict[str, Material] = _load()

_Inventory = typing.TypeVar("_Inventory", bound="Inventory")


class Inventory(collections.UserDict):
    """A mapping of every available material with an associated quantity."""

    def __init__(
        self,
        items: typing.Optional[typing.Union[typing.Dict[str, int], "Inventory"]] = None,
    ) -> None:
        self.data = {name: 0 for name in materials}
        if items:
            self.update(items)

    def __setitem__(self, key, value) -> None:
        normalized_key = key.lower()
        if normalized_key not in materials:
            raise KeyError(key)
        if value < 0:
            raise ValueError(value)
        super().__setitem__(normalized_key, value)

    def __getitem__(self, key: str) -> int:
        normalized_key = key.lower()
        if normalized_key not in materials:
            raise KeyError(key)
        return super().__getitem__(normalized_key)

    def has(self, items: typing.Union[typing.Dict[str, int], "Inventory"]) -> bool:
        return all([self[name] >= quantity for name, quantity in items.items()])

    def __add__(self, other: "Inventory") -> "Inventory":
        return Inventory({name: self[name] + other[name] for name in materials})

    def __sub__(self, other: "Inventory") -> "Inventory":
        return Inventory({name: self[name] - other[name] for name in materials})

    def trades(self, symbol: str, quantity: int) -> "Materials":
        if self.has(Materials({symbol: quantity})):
            return Materials({})
        raise ValueError

    def trade_for(self, other: "Inventory") -> "Inventory":
        def accumulate(inv: "Inventory", mat: typing.Tuple[str, int]) -> "Inventory":
            symbol, quantity = mat
            return inv - inv.trades(symbol, quantity)

        return functools.reduce(accumulate, other.items(), self)

    @classmethod
    def from_event(
        cls: typing.Type[_Inventory], event: typing.Dict[str, typing.Any]
    ) -> _Inventory:
        if event.get("event") != "Materials":
            raise ValueError("Not a 'Materials' event")
        return cls(
            {
                material["Name"]: material["Count"]
                for material in [
                    *event.get("Raw", []),
                    *event.get("Manufactured", []),
                    *event.get("Encoded", []),
                ]
            }
        )


class Materials(Inventory):
    """Same as Inventory, but omits zeroed materials."""

    def __init__(
        self,
        items: typing.Optional[typing.Union[typing.Dict[str, int], "Inventory"]] = None,
    ) -> None:
        self.data: typing.Dict[str, int] = {}
        if items:
            self.update(items)

    def __setitem__(self, key: str, value: int) -> None:
        normalized_key = key.lower()
        if normalized_key not in materials:
            raise KeyError(key)
        if value < 0:
            raise ValueError(value)
        if value == 0:
            return
        super().__setitem__(normalized_key, value)

    def __getitem__(self, key: str) -> int:
        normalized_key = key.lower()
        if normalized_key not in materials:
            raise KeyError(key)
        return self.data.get(normalized_key.lower(), 0)

    def __add__(self, other: "Inventory") -> "Materials":
        return Materials({name: self[name] + other[name] for name in materials})

    def __sub__(self, other: "Inventory") -> "Materials":
        return Materials({name: self[name] - other[name] for name in materials})
