import fractions
import json
import pathlib
import typing
import unittest

from elite_engineering import materials


class TradeCalculationTests(unittest.TestCase):
    def format_ratio(self, ratio: typing.Optional[fractions.Fraction]) -> str:
        if not ratio:
            return "-"
        return "{} → {}".format(ratio.numerator, ratio.denominator)

    def test_same_category_exchange(self) -> None:
        expected = [
            ["1 → 1", "1 → 3", "1 → 9", "1 → 27", "1 → 81"],
            ["6 → 1", "1 → 1", "1 → 3", "1 → 9", "1 → 27"],
            ["36 → 1", "6 → 1", "1 → 1", "1 → 3", "1 → 9"],
            ["216 → 1", "36 → 1", "6 → 1", "1 → 1", "1 → 3"],
            ["-", "216 → 1", "36 → 1", "6 → 1", "1 → 1"],
        ]

        actual = [
            [
                self.format_ratio(
                    materials.Material(
                        "symbol", "name", "raw", "category", grade_in
                    ).trade_ratio(
                        materials.Material(
                            "symbol", "name", "raw", "category", grade_out
                        )
                    )
                )
                for grade_in in [1, 2, 3, 4, 5]
            ]
            for grade_out in [1, 2, 3, 4, 5]
        ]
        self.assertEqual(expected, actual)

    def test_different_category_exchange(self) -> None:
        expected = [
            ["6 → 1", "2 → 1", "2 → 3", "2 → 9", "2 → 27"],
            ["36 → 1", "6 → 1", "2 → 1", "2 → 3", "2 → 9"],
            ["216 → 1", "36 → 1", "6 → 1", "2 → 1", "2 → 3"],
            ["-", "216 → 1", "36 → 1", "6 → 1", "2 → 1"],
            ["-", "-", "216 → 1", "36 → 1", "6 → 1"],
        ]

        actual = [
            [
                self.format_ratio(
                    materials.Material(
                        "symbol", "name", "raw", "category", grade_in
                    ).trade_ratio(
                        materials.Material("symbol", "name", "raw", "other", grade_out)
                    )
                )
                for grade_in in [1, 2, 3, 4, 5]
            ]
            for grade_out in [1, 2, 3, 4, 5]
        ]

        self.assertEqual(expected, actual)

    def test_different_type_exchange(self) -> None:
        expected = [
            ["-", "-", "-", "-", "-"],
            ["-", "-", "-", "-", "-"],
            ["-", "-", "-", "-", "-"],
            ["-", "-", "-", "-", "-"],
            ["-", "-", "-", "-", "-"],
        ]

        actual = [
            [
                self.format_ratio(
                    materials.Material(
                        "symbol", "name", "raw", "category", grade_in
                    ).trade_ratio(
                        materials.Material(
                            "symbol", "name", "encoded", "other", grade_out
                        )
                    )
                )
                for grade_in in [1, 2, 3, 4, 5]
            ]
            for grade_out in [1, 2, 3, 4, 5]
        ]

        self.assertEqual(expected, actual)


class InventoryTests(unittest.TestCase):
    def make(
        self, items: typing.Optional[typing.Dict[str, int]] = None
    ) -> "materials.Inventory":
        return materials.Inventory(items)

    def from_event(self, event: typing.Dict[str, typing.Any]) -> "materials.Inventory":
        return materials.Inventory.from_event(event)

    def test_create_empty(self) -> None:
        inventory = self.make()
        self.assertEqual(set(materials.materials.keys()), set(inventory.keys()))

    def test_create_populated(self) -> None:
        inventory = self.make({"Carbon": 5, "ShieldEmitters": 3})
        self.assertEqual(5, inventory["Carbon"])
        self.assertEqual(3, inventory["ShieldEmitters"])
        self.assertEqual(8, sum(inventory.values()))

    def test_create_with_invalid_materials(self) -> None:
        with self.assertRaises(KeyError) as exc:
            self.make({"Carbon": 5, "Pizza": 12})
        self.assertEqual(exc.exception.args, ("Pizza",))

    def test_create_with_invalid_quantities(self) -> None:
        with self.assertRaises(ValueError) as exc:
            self.make({"Carbon": -5})
        self.assertEqual(exc.exception.args, (-5,))

    def test_has_materials(self) -> None:
        inventory = self.make({"Carbon": 5, "ShieldEmitters": 3})
        self.assertTrue(inventory.has({"Carbon": 1, "ShieldEmitters": 3}))

    def test_lacks_materials(self) -> None:
        inventory = self.make({"Carbon": 5, "ShieldEmitters": 3})
        self.assertTrue(not inventory.has({"Nickel": 1}))
        self.assertTrue(not inventory.has({"Carbon": 6, "ShieldEmitters": 3}))

    def test_has_raises_on_invalid_materials(self) -> None:
        inventory = self.make()
        with self.assertRaises(KeyError) as exc:
            inventory.has({"Carbon": 5, "Pizza": 12})
        self.assertEqual(exc.exception.args, ("Pizza",))

    def test_addition(self) -> None:
        a = self.make({"Carbon": 5, "ShieldEmitters": 3})
        b = self.make({"Iron": 2, "ShieldEmitters": 3})
        expected = self.make({"Carbon": 5, "Iron": 2, "ShieldEmitters": 6})
        self.assertEqual(expected, a + b)

    def test_subtraction(self) -> None:
        a = self.make({"Carbon": 5, "Iron": 2, "ShieldEmitters": 6})
        b = self.make({"Carbon": 5, "ShieldEmitters": 3})
        expected = self.make({"Iron": 2, "ShieldEmitters": 3})
        self.assertEqual(expected, a - b)

    def test_subtraction_cannot_result_in_negative_quantities(self) -> None:
        a = self.make({"Carbon": 5, "Iron": 2, "ShieldEmitters": 6})
        b = self.make({"Carbon": 6, "ShieldEmitters": 3})
        with self.assertRaises(ValueError) as exc:
            a - b
        self.assertEqual(exc.exception.args, (-1,))

    def test_load_from_materials_event(self) -> None:
        event = {
            "timestamp": "2022-01-23T02:58:12Z",
            "event": "Materials",
            "Raw": [
                {"Name": "manganese", "Count": 127},
                {"Name": "iron", "Count": 37},
                {"Name": "phosphorus", "Count": 2},
                {"Name": "nickel", "Count": 5},
                {"Name": "lead", "Count": 69},
                {"Name": "germanium", "Count": 12},
            ],
            "Manufactured": [
                {
                    "Name": "salvagedalloys",
                    "Name_Localised": "Salvaged Alloys",
                    "Count": 1,
                },
                {
                    "Name": "crystalshards",
                    "Name_Localised": "Crystal Shards",
                    "Count": 36,
                },
                {
                    "Name": "shieldemitters",
                    "Name_Localised": "Shield Emitters",
                    "Count": 36,
                },
            ],
            "Encoded": [
                {
                    "Name": "bulkscandata",
                    "Name_Localised": "Anomalous Bulk Scan Data",
                    "Count": 28,
                },
                {
                    "Name": "shieldcyclerecordings",
                    "Name_Localised": "Distorted Shield Cycle Recordings",
                    "Count": 66,
                },
                {
                    "Name": "archivedemissiondata",
                    "Name_Localised": "Irregular Emission Data",
                    "Count": 6,
                },
            ],
        }

        self.assertEqual(
            self.make(
                {
                    "manganese": 127,
                    "iron": 37,
                    "phosphorus": 2,
                    "nickel": 5,
                    "lead": 69,
                    "germanium": 12,
                    "salvagedalloys": 1,
                    "crystalshards": 36,
                    "shieldemitters": 36,
                    "bulkscandata": 28,
                    "shieldcyclerecordings": 66,
                    "archivedemissiondata": 6,
                }
            ),
            self.from_event(event),
        )

    def test_load_from_invalid_event(self) -> None:
        event: typing.Dict[str, typing.Any] = {}
        with self.assertRaises(ValueError) as exc:
            self.from_event(event)
        self.assertEqual(exc.exception.args, ("Not a 'Materials' event",))

    def test_trade_unnecessary(self) -> None:
        inventory = self.make({"carbon": 3, "niobium": 1})
        recipe = materials.Materials({"carbon": 3})
        self.assertEqual(inventory, inventory.trade_for(recipe))

    def test_trade_impossible(self) -> None:
        inventory = self.make()
        recipe = materials.Materials({"carbon": 3})
        with self.assertRaises(ValueError):
            inventory.trade_for(recipe)

    @unittest.skip("Not implemented")
    def test_trade_grade_2_for_grade_1(self) -> None:
        inventory = self.make({"vanadium": 1, "niobium": 1})
        recipe = materials.Materials({"carbon": 3})
        self.assertEqual(
            self.make({"carbon": 3, "niobium": 1}), inventory.trade_for(recipe)
        )


class MaterialsTests(InventoryTests):
    def make(
        self, items: typing.Optional[typing.Dict[str, int]] = None
    ) -> "materials.Materials":
        return materials.Materials(items)

    def from_event(self, event: typing.Dict[str, typing.Any]) -> "materials.Materials":
        return materials.Materials.from_event(event)

    def test_create_empty(self) -> None:
        empty = self.make()
        self.assertEqual(set(), set(empty.keys()))
